@@ -23,6 +23,7 @@
 -- from the untyped 'Map' representation obtained from 'decodeWire'.
 
 {-# LANGUAGE BangPatterns               #-}
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE DeriveFunctor              #-}
 {-# LANGUAGE LambdaCase                 #-}
 {-# LANGUAGE OverloadedStrings          #-}
@@ -53,7 +54,9 @@ module Proto3.Wire.Decode
     , enum
     , byteString
     , lazyByteString
+    , shortByteString
     , text
+    , shortText
     , packedVarints
     , packedFixed32
     , packedFixed64
@@ -74,8 +77,6 @@ module Proto3.Wire.Decode
     , embedded'
       -- * ZigZag codec
     , zigZagDecode
-      -- * Exported For Doctest Only
-    , toMap
     ) where
 
 import           Control.Applicative
@@ -85,16 +86,19 @@ import           Control.Monad           ( msum, foldM )
 import           Data.Bits
 import qualified Data.ByteString         as B
 import qualified Data.ByteString.Lazy    as BL
+import qualified Data.ByteString.Short   as BS
+#if !MIN_VERSION_base(4,20,0)
 import           Data.Foldable           ( foldl' )
+#endif
 import qualified Data.IntMap.Strict      as M -- TODO intmap
 import           Data.Maybe              ( fromMaybe )
-import           Data.Monoid             ( (<>) )
 import           Data.Serialize.Get      ( Get, getWord8, getInt32le
                                          , getInt64le, getWord32le, getWord64le
                                          , runGet )
 import           Data.Serialize.IEEE754  ( getFloat32le, getFloat64le )
 import           Data.Text.Lazy          ( Text, pack )
 import           Data.Text.Lazy.Encoding ( decodeUtf8' )
+import qualified Data.Text.Short         as Text.Short
 import qualified Data.Traversable        as T
 import           Data.Int                ( Int32, Int64 )
 import           Data.Word               ( Word8, Word32, Word64 )
@@ -124,29 +128,48 @@ data ParsedField = VarintField Word64
 -- | Convert key-value pairs to a map of keys to a sequence of values with that
 -- key, in their reverse occurrence order.
 --
--- >>> toMap ([(FieldNumber 1, 3),(FieldNumber 2, 4),(FieldNumber 1, 6)] :: [(FieldNumber,Int)])
--- fromList [(1,[6,3]),(2,[4])]
 --
-toMap :: [(FieldNumber, v)] -> M.IntMap [v]
-toMap kvs0 = M.fromListWith (<>) . map (fmap (:[])) . map (first (fromIntegral . getFieldNumber)) $ kvs0
+decodeWireMessage :: B.ByteString -> Either String RawMessage
+decodeWireMessage = decodeWire0 combineSeen' Nothing close
+  where
+    close Nothing = M.empty
+    close (Just (m, k, v)) = M.insertWith (++) k v m
+
+    combineSeen' :: Maybe (M.IntMap [v], Int, [v]) -> FieldNumber -> v -> Maybe (M.IntMap [v], Int, [v])
+    combineSeen' b (FieldNumber fn) v = combineSeen b (fromIntegral fn) v
+
+    combineSeen :: Maybe (M.IntMap [v], Int, [v]) -> Int -> v -> Maybe (M.IntMap [v], Int, [v])
+    combineSeen Nothing k1 a1 = Just (M.empty, k1, [a1])
+    combineSeen (Just (m, k2, as)) k1 a1 =
+      if k1 == k2
+        then Just (m, k1, a1 : as)
+        -- It might seem that we want to use DList but we don't because:
+        -- - alter has worse performance than insertWith, and there's no upsert
+        -- - We're building up a list of elements in a recursive way
+        --    that will be opaque to GHC
+        -- - DList would add another dependency
+        else let !m' = M.insertWith (++) k2 as m
+             in Just (m', k1, [a1])
+
+decodeWire :: B.ByteString -> Either String [(FieldNumber, ParsedField)]
+decodeWire = decodeWire0 (\xs k v -> (k,v):xs) [] reverse
 
 -- | Parses data in the raw wire format into an untyped 'Map' representation.
-decodeWire :: B.ByteString -> Either String [(FieldNumber, ParsedField)]
-decodeWire bstr = drloop bstr []
+decodeWire0 :: (b -> FieldNumber -> ParsedField -> b) -> b -> (b -> r) -> B.ByteString -> Either String r
+decodeWire0 cl z finish bstr = drloop bstr z
  where
-   drloop !bs xs | B.null bs = Right $ reverse xs
+   drloop !bs xs | B.null bs = Right $ finish xs
    drloop !bs xs | otherwise = do
       (w, rest) <- takeVarInt bs
       wt <- gwireType $ fromIntegral (w .&. 7)
       let fn = w `shiftR` 3
       (res, rest2) <- takeWT wt rest
-      drloop rest2 ((FieldNumber fn,res):xs)
-
+      drloop rest2 (cl xs (FieldNumber fn) res)
+{-# INLINE decodeWire0 #-}
 
 eitherUncons :: B.ByteString -> Either String (Word8, B.ByteString)
 eitherUncons = maybe (Left "failed to parse varint128") Right . B.uncons
 
-
 takeVarInt :: B.ByteString -> Either String (Word64, B.ByteString)
 takeVarInt !bs =
   case B.uncons bs of
@@ -191,6 +214,7 @@ takeVarInt !bs =
                 if w10 < 128 then return (val9 + (fromIntegral w10 `shiftL` 63), r10) else do
 
                  Left ("failed to parse varint128: too big; " ++ show val6)
+{-# INLINE takeVarInt #-}
 
 
 gwireType :: Word8 -> Either String WireType
@@ -199,6 +223,7 @@ gwireType 5 = return Fixed32
 gwireType 1 = return Fixed64
 gwireType 2 = return LengthDelimited
 gwireType wt = Left $ "wireType got unknown wire type: " ++ show wt
+{-# INLINE gwireType #-}
 
 safeSplit :: Int -> B.ByteString -> Either String (B.ByteString, B.ByteString)
 safeSplit !i !b | B.length b < i = Left "failed to parse varint128: not enough bytes"
@@ -211,6 +236,7 @@ takeWT Fixed64 !b = fmap (first Fixed64Field) $ safeSplit 8 b
 takeWT LengthDelimited b = do
    (!len, rest) <- takeVarInt b
    fmap (first LengthDelimitedField) $ safeSplit (fromIntegral len) rest
+{-# INLINE takeWT #-}
 
 
 -- * Parser Interface
@@ -290,9 +316,10 @@ foldFields parsers = foldM applyOne
 -- | Parse a message (encoded in the raw wire format) using the specified
 -- `Parser`.
 parse :: Parser RawMessage a -> B.ByteString -> Either ParseError a
-parse parser bs = case decodeWire bs of
+parse parser bs = case decodeWireMessage bs of
     Left err -> Left (BinaryError (pack err))
-    Right res -> runParser parser (toMap res)
+    Right res -> runParser parser res
+{-# INLINE parse #-}
 
 -- | To comply with the protobuf spec, if there are multiple fields with the same
 -- field number, this will always return the last one.
@@ -301,22 +328,31 @@ parsedField xs = case xs of
     [] -> Nothing
     (x:_) -> Just x
 
-throwWireTypeError :: Show input
-                   => String
-                   -> input
-                   -> Either ParseError expected
-throwWireTypeError expected wrong =
-    Left (WireTypeError (pack msg))
+wireTypeError :: String
+              -> RawPrimitive
+              -> ParseError
+wireTypeError expected wrong = WireTypeError (pack msg)
   where
     msg = "Wrong wiretype. Expected " ++ expected ++ " but got " ++ show wrong
 
-throwCerealError :: String -> String -> Either ParseError a
-throwCerealError expected cerealErr =
-    Left (BinaryError (pack msg))
+throwWireTypeError :: String
+                   -> RawPrimitive
+                   -> Either ParseError expected
+throwWireTypeError expected wrong =
+    Left (wireTypeError expected wrong)
+{-# INLINE throwWireTypeError #-}
+
+cerealTypeError :: String -> String -> ParseError
+cerealTypeError expected cerealErr = (BinaryError (pack msg))
   where
     msg = "Failed to parse contents of " ++
         expected ++ " field. " ++ "Error from cereal was: " ++ cerealErr
 
+throwCerealError :: String -> String -> Either ParseError a
+throwCerealError expected cerealErr =
+    Left (cerealTypeError expected cerealErr)
+{-# INLINE throwCerealError #-}
+
 parseVarInt :: Integral a => Parser RawPrimitive a
 parseVarInt = Parser $
     \case
@@ -351,8 +387,7 @@ runGetFixed64 g = Parser $
 bytes :: Parser RawPrimitive B.ByteString
 bytes = Parser $
     \case
-        LengthDelimitedField bs ->
-            return $! B.copy bs
+        LengthDelimitedField bs -> return $! B.copy bs
         wrong -> throwWireTypeError "bytes" wrong
 
 -- | Parse a Boolean value.
@@ -394,6 +429,13 @@ byteString = bytes
 lazyByteString :: Parser RawPrimitive BL.ByteString
 lazyByteString = fmap BL.fromStrict bytes
 
+-- | Parse a primitive with the @bytes@ wire type as a 'BS.ShortByteString'.
+shortByteString :: Parser RawPrimitive BS.ShortByteString
+shortByteString = Parser $
+    \case
+        LengthDelimitedField bs -> return $! BS.toShort bs
+        wrong -> throwWireTypeError "bytes" wrong
+
 -- | Parse a primitive with the @bytes@ wire type as 'Text'.
 text :: Parser RawPrimitive Text
 text = Parser $
@@ -405,6 +447,16 @@ text = Parser $
                 Right txt -> return txt
         wrong -> throwWireTypeError "string" wrong
 
+-- | Parse a primitive with the @bytes@ wire type as `Text.Short.ShortText`.
+shortText :: Parser RawPrimitive Text.Short.ShortText
+shortText = Parser $
+    \case
+        LengthDelimitedField bs ->
+            case Text.Short.fromByteString bs of
+                Nothing -> Left (BinaryError (pack ("Failed to decode UTF-8")))
+                Just txt -> return txt
+        wrong -> throwWireTypeError "string" wrong
+
 -- | Parse a primitive with an enumerated type.
 --
 -- This parser will return 'Left' if the encoded integer value
@@ -487,6 +539,7 @@ sfixed64 = runGetFixed64 getInt64le
 -- > one float `at` fieldNumber 1 :: Parser RawMessage (Maybe Float)
 at :: Parser RawField a -> FieldNumber -> Parser RawMessage a
 at parser fn = Parser $ runParser parser . fromMaybe mempty . M.lookup (fromIntegral . getFieldNumber $ fn)
+{-# INLINE at #-}
 
 -- | Try to parse different field numbers with their respective parsers. This is
 -- used to express alternative between possible fields of a oneof.
@@ -539,15 +592,20 @@ one parser def = Parser (fmap (fromMaybe def) . traverse (runParser parser) . pa
 repeated :: Parser RawPrimitive a -> Parser RawField [a]
 repeated parser = Parser $ fmap reverse . mapM (runParser parser)
 
+
+embeddedParseError :: String -> ParseError
+embeddedParseError err = EmbeddedError msg Nothing
+  where
+    msg = "Failed to parse embedded message: " <> (pack err)
+{-# NOINLINE embeddedParseError #-}
+
 -- | For a field containing an embedded message, parse as far as getting the
 -- wire-level fields out of the message.
 embeddedToParsedFields :: RawPrimitive -> Either ParseError RawMessage
 embeddedToParsedFields (LengthDelimitedField bs) =
-    case decodeWire bs of
-        Left err -> Left (EmbeddedError ("Failed to parse embedded message: "
-                                             <> (pack err))
-                                        Nothing)
-        Right result -> return (toMap result)
+    case decodeWireMessage bs of
+        Left err -> Left (embeddedParseError err)
+        Right result -> Right result
 embeddedToParsedFields wrong =
     throwWireTypeError "embedded" wrong
 
@@ -569,6 +627,7 @@ embedded p = Parser $
                let combinedMap = foldl' (M.unionWith (<>)) M.empty innerMaps
                parsed <- runParser p combinedMap
                return $ Just parsed
+{-# INLINE embedded #-}
 
 -- | Create a primitive parser for an embedded message from a message parser.
 --
@@ -579,10 +638,11 @@ embedded' parser = Parser $
     \case
         LengthDelimitedField bs ->
             case parse parser bs of
-                Left err -> Left (EmbeddedError "Failed to parse embedded message."
-                                                (Just err))
+                Left err -> Left (EmbeddedError "Failed to parse embedded message." (Just err))
                 Right result -> return result
         wrong -> throwWireTypeError "embedded" wrong
+{-# INLINE embedded' #-}
+
 
 
 -- TODO test repeated and embedded better for reverse logic...
