@@ -25,6 +25,7 @@
 -- [6,0,0,0,42,206,187]
 
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE CPP #-}
 
 module Proto3.Wire.Reverse
     ( -- * `BuildR` type
@@ -63,6 +64,7 @@ module Proto3.Wire.Reverse
     , stringUtf8
     , textUtf8
     , lazyTextUtf8
+    , shortTextUtf8
     , wordBase128LEVar
     , wordBase128LEVar_inline
     , word32Base128LEVar
@@ -84,7 +86,6 @@ module Proto3.Wire.Reverse
 
 import           Data.Bits                     ( (.&.) )
 import qualified Data.ByteString               as B
-import qualified Data.ByteString.Internal      as BI
 import qualified Data.ByteString.Lazy          as BL
 import qualified Data.ByteString.Lazy.Internal as BLI
 import qualified Data.ByteString.Short         as BS
@@ -94,14 +95,21 @@ import           Data.Char                     ( ord )
 import           Data.Int                      ( Int8, Int16, Int32, Int64 )
 import qualified Data.Text                     as T
 import qualified Data.Text.Internal            as TI
-import qualified Data.Text.Internal.Fusion     as TIF
 import qualified Data.Text.Lazy                as TL
+import qualified Data.Text.Short               as TS
 import           Data.Vector.Generic           ( Vector )
 import           Data.Word                     ( Word8, Word16, Word32, Word64 )
-import           Foreign                       ( castPtr )
+import           Foreign                       ( castPtr, copyBytes )
 import           Proto3.Wire.Reverse.Internal
 import qualified Proto3.Wire.Reverse.Prim      as Prim
 
+#if MIN_VERSION_text(2,0,0)
+import           Control.Monad.ST.Unsafe       (unsafeSTToIO)
+import qualified Data.Text.Array               as TA
+#else
+import qualified Data.Text.Internal.Fusion     as TIF
+#endif
+
 -- $setup
 -- >>> :set -XOverloadedStrings
 -- >>> :module Proto3.Wire.Reverse
@@ -132,7 +140,7 @@ byteString bs = withUnused $ \unused ->
     then
       unsafeConsume len $ \dst ->
         BU.unsafeUseAsCString bs $ \src ->
-          BI.memcpy dst (castPtr src) len
+          copyBytes dst (castPtr src) len
     else
       prependChunk bs
 
@@ -168,7 +176,7 @@ lazyByteString = etaBuildR $ scan (ReverseChunks BL.empty)
           (prepend (ReverseChunks cs) <>) $
             unsafeConsume len $ \dst ->
               BU.unsafeUseAsCString c $ \src ->
-                BI.memcpy dst (castPtr src) len
+                copyBytes dst (castPtr src) len
         else
           prependReverseChunks (ReverseChunks(BLI.Chunk c cs))
 
@@ -188,7 +196,7 @@ shortByteString bs = withUnused $ \unused ->
         writeChunk bs 0 len
       else
         let rest = len - unused in
-        writeChunk bs unused rest <> reallocate rest <> writeChunk bs 0 unused
+        writeChunk bs 0 rest <> reallocate rest <> writeChunk bs rest unused
   where
     writeChunk src off len =
       unsafeConsume len $ \dst ->
@@ -506,7 +514,26 @@ stringUtf8 = foldMap charUtf8
 -- >>> textUtf8 "←↑→↓"
 -- Proto3.Wire.Reverse.lazyByteString "\226\134\144\226\134\145\226\134\146\226\134\147"
 textUtf8 :: T.Text -> BuildR
+#if MIN_VERSION_text(2,0,0)
+textUtf8 = etaBuildR $ \(TI.Text arr off word8Count) ->
+  -- For version 2 of the "text" package, the in-memory
+  -- representation is UTF-8.  We can just write it out.
+  withUnused $ \unused ->
+      if word8Count <= unused
+        then
+          writeChunk arr off word8Count
+        else
+          let rest = word8Count - unused in
+          writeChunk arr off rest <>
+            reallocate rest <> writeChunk arr (off + rest) unused
+    where
+      writeChunk src off len =
+        unsafeConsume len $ \dst -> unsafeSTToIO $
+          TA.copyToPointer src off dst len
+#else
 textUtf8 = etaBuildR $ \txt@(TI.Text _ _ word16Count) ->
+  -- For version 1 of the "text" package, the in-memory
+  -- representation is UTF-16.  We must transcode to UTF-8.
   case TIF.reverseStream txt of
     TIF.Stream next t0 _ -> ensure bound (go t0)
       where
@@ -531,6 +558,7 @@ textUtf8 = etaBuildR $ \txt@(TI.Text _ _ word16Count) ->
           TIF.Skip t2 -> go t2
           TIF.Yield !ch t2 ->
             go t2 <> Prim.unsafeBuildBoundedPrim (Prim.charUtf8 ch)
+#endif
 
 -- | Convert a Unicode lazy `TL.Text` to a `BuildR` using a @UTF-8@ encoding
 --
@@ -545,6 +573,19 @@ textUtf8 = etaBuildR $ \txt@(TI.Text _ _ word16Count) ->
 lazyTextUtf8 :: TL.Text -> BuildR
 lazyTextUtf8 = TL.foldrChunks ((<>) . textUtf8) mempty
 
+-- | Convert a `TS.ShortText` to a `BuildR` using a @UTF-8@ encoding.
+--
+-- > shortTextUtf8 (x <> y) = shortTextUtf8 x <> shortTextUtf8 y
+-- >
+-- > shortTextUtf8 mempty = mempty
+--
+-- >>> shortTextUtf8 "ABC"
+-- Proto3.Wire.Reverse.lazyByteString "ABC"
+-- >>> shortTextUtf8 "←↑→↓"
+-- Proto3.Wire.Reverse.lazyByteString "\226\134\144\226\134\145\226\134\146\226\134\147"
+shortTextUtf8 :: TS.ShortText -> BuildR
+shortTextUtf8 = shortByteString . TS.toShortByteString
+
 -- | Convert a `Word` to a `BuildR` using this variable-length encoding:
 --
 --   1. Convert the given value to a base 128 representation
