@@ -23,15 +23,26 @@
 {-# LANGUAGE UnboxedTuples #-}
 
 module Proto3.Wire.Reverse.Internal
-    ( BuildR(..)
+    ( BuildR (..)
+    , BuildRState (..)
     , appendBuildR
     , foldlRVector
     , toBuildR
     , fromBuildR
     , etaBuildR
     , runBuildR
+    , SealedState (SealedState, sealedSB, totalSB, stateVarSB, statePtrSB, recycledSB)
+    , sealBuffer
+    , smallChunkSize
+    , readState
+    , readSpace
+    , writeState
+    , writeSpace
+    , metaDataSize
+    , metaDataAlign
     , withUnused
     , withTotal
+    , readTotal
     , withLengthOf
     , withLengthOf#
     , reallocate
@@ -134,7 +145,11 @@ instance Semigroup BuildR
 instance Monoid BuildR
   where
     mempty = BuildR (\v u s -> (# v, u, s #))
+#ifdef ghcjs_HOST_OS
+    {-# NOINLINE mempty #-}
+#else
     {-# INLINE mempty #-}
+#endif
 
     mappend = (<>)
     {-# INLINE mappend #-}
@@ -447,7 +462,7 @@ sealBuffer# addr unused s0 =
       -- until a state action frees the stable pointer or modifies the state
       -- variable, the stable pointer will reference the state variable,
       -- which in turn will reference the current buffer.
-      let allocation = P.sizeofMutableByteArray buffer - metaDataSize
+      allocation <- subtract metaDataSize <$> P.getSizeofMutableByteArray buffer
       if allocation <= I# unused
         then
           pure (oldSealed, total, stateVar, statePtr, Just buffer)
@@ -558,8 +573,8 @@ afterPrependChunks# SealedState
       Just buf -> do
         -- Recycle the old current buffer, from which
         -- we already copied what we wished to keep.
-        let u1 = P.sizeofMutableByteArray buf - metaDataSize
-            !(PTR base) = P.mutableByteArrayContents buf
+        u1 <- subtract metaDataSize <$> P.getSizeofMutableByteArray buf
+        let !(PTR base) = P.mutableByteArrayContents buf
             !v1 = plusPtr (Ptr base) (metaDataSize + u1)
             !m = plusPtr (Ptr base) metaDataSize
         writeSpace m (u1 + total)
