@@ -16,24 +16,28 @@
 
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 {-# OPTIONS_GHC -Wno-warnings-deprecations #-}
 
 module Main where
 
-import           Control.Arrow         ( (&&&), first, second )
+import           Control.Arrow         ( (&&&), second )
 import           Control.Monad         ( guard, void )
 import           Control.Monad.Trans.State ( StateT(..) )
 import qualified Data.Bits             as Bits
 import qualified Data.ByteString       as B
+import qualified Data.ByteString.Builder.Extra as BB
 import qualified Data.ByteString.Lazy  as BL
+import qualified Data.ByteString.Short as BS
 import qualified Data.ByteString.Builder.Internal as BBI
 import           Data.Either           ( isLeft )
 import           Data.Maybe            ( fromMaybe )
-import           Data.Monoid           ( (<>) )
 import           Data.Int
-import           Data.List             ( group )
+import           Data.List             ( sort )
+import qualified Data.List.NonEmpty    as NE
 import qualified Data.Text.Lazy        as T
+import qualified Data.Text.Short       as TS
 import qualified Data.Vector           as V
 import           Data.Word             ( Word8, Word64 )
 import           Foreign               ( sizeOf )
@@ -41,14 +45,12 @@ import           Foreign               ( sizeOf )
 import           Proto3.Wire
 import qualified Proto3.Wire.Builder   as Builder
 import qualified Proto3.Wire.Reverse   as Reverse
-import qualified Proto3.Wire.Reverse.Prim as Prim
 import qualified Proto3.Wire.Encode    as Encode
 import qualified Proto3.Wire.Decode    as Decode
 
 import qualified Test.DocTest
 import           Test.QuickCheck       ( (===), Arbitrary )
 import           Test.Tasty
-import           Test.Tasty.HUnit      ( (@=?) )
 import qualified Test.Tasty.HUnit      as HU
 import qualified Test.Tasty.QuickCheck as QC
 
@@ -57,6 +59,7 @@ main = do
     Test.DocTest.doctest
       [ "-isrc"
       , "-fobject-code"
+      , "-Wno-warnings-deprecations"
       , "src/Proto3/Wire/Builder.hs"
       , "src/Proto3/Wire/Reverse.hs"
       , "src/Proto3/Wire/Encode.hs"
@@ -73,6 +76,7 @@ tests = testGroup "Tests" [ roundTripTests
                           , decodeNonsense
                           , varIntHeavyTests
                           , packedLargeTests
+                          , decodeWireRoundTrip
                           ]
 
 data StringOrInt64 = TString T.Text | TInt64 Int64
@@ -133,6 +137,25 @@ roundTripTests = testGroup "Roundtrip tests"
                                        (Encode.text (fieldNumber 1) . T.pack)
                                        (one (fmap T.unpack Decode.text) mempty `at`
                                             fieldNumber 1)
+                           , roundTrip "shortText"
+                                       (Encode.shortText (fieldNumber 1) . TS.pack)
+                                       (one (fmap TS.unpack Decode.shortText) mempty `at`
+                                            fieldNumber 1)
+                           , roundTripFor (QC.oneof [QC.arbitrary, genManyOctets])
+                                       "byteString"
+                                       (Encode.byteString (fieldNumber 1) . B.pack)
+                                       (one (fmap B.unpack Decode.byteString) mempty `at`
+                                            fieldNumber 1)
+                           , roundTripFor genLazyByteString
+                                       "lazyByteString"
+                                       (Encode.lazyByteString (fieldNumber 1))
+                                       (one (Decode.lazyByteString) mempty `at`
+                                            fieldNumber 1)
+                           , roundTripFor (QC.oneof [QC.arbitrary, genManyOctets])
+                                       "shortByteString"
+                                       (Encode.shortByteString (fieldNumber 1) . BS.pack)
+                                       (one (fmap BS.unpack Decode.shortByteString) mempty `at`
+                                            fieldNumber 1)
                            , roundTrip "embedded"
                                        (Encode.embedded (fieldNumber 1) .
                                             Encode.int32 (fieldNumber 1))
@@ -218,7 +241,6 @@ roundTripTests = testGroup "Roundtrip tests"
                                                , (fieldNumber 3, Just . TString <$> one Decode.text mempty)
                                                ]
                                         )
-
                            ]
 
 roundTrip :: (Show a, Eq a, Arbitrary a)
@@ -226,14 +248,45 @@ roundTrip :: (Show a, Eq a, Arbitrary a)
           -> (a -> Encode.MessageBuilder)
           -> Decode.Parser Decode.RawMessage a
           -> TestTree
-roundTrip name encode decode =
-    QC.testProperty name $
-        \x -> do
-            let bytes = Encode.toLazyByteString (encode x)
+roundTrip = roundTripFor QC.arbitrary
+
+roundTripFor :: (Show a, Eq a)
+             => QC.Gen a
+             -> String
+             -> (a -> Encode.MessageBuilder)
+             -> Decode.Parser Decode.RawMessage a
+             -> TestTree
+roundTripFor gen name encode decode =
+    QC.testProperty name $ QC.forAll gen $
+        \x ->
+            let bytes = Encode.toLazyByteString (encode x) in
             case Decode.parse decode (BL.toStrict bytes) of
                 Left _ -> error "Could not decode encoded message"
                 Right x' -> x === x'
 
+genManyOctets :: QC.Gen [Word8]
+genManyOctets =
+  QC.vector =<< QC.choose (BB.smallChunkSize - 64, BB.smallChunkSize + 64)
+
+genLazyByteString :: QC.Gen BL.ByteString
+genLazyByteString = do
+  octets <- genManyOctets
+  let total = length octets
+  splits <- QC.listOf (QC.choose (0, total))
+  let go :: Int -> [Int] -> [Word8] -> [[Word8]]
+      go x [] os = [take (total - x) os]
+      go x (y : ys) os = let (o1, o2) = splitAt (y - x) os in o1 : go y ys o2
+  pure $ BL.fromChunks $ map B.pack $ go 0 (sort splits) octets
+
+decodeWireRoundTrip :: TestTree
+decodeWireRoundTrip = QC.testProperty "decodeWire round trips" $
+  \(inp :: [(FieldNumber, Int32)]) ->
+    let bytes = Encode.toLazyByteString (foldMap (\(k, v) -> Encode.int32 k v) inp)
+        x = map (second $ Decode.VarintField . fromIntegral) inp
+    in case Decode.decodeWire (BL.toStrict bytes) of
+          Left _ -> error "decodeWire failed"
+          Right x' -> x === x'
+
 buildSingleChunk :: TestTree
 buildSingleChunk = HU.testCase "Legacy Builder creates a single chunk" $ do
   let chunks = length . BL.toChunks . Builder.toLazyByteString
@@ -249,9 +302,9 @@ buildSingleChunk = HU.testCase "Legacy Builder creates a single chunk" $ do
 
 parseBytes :: Int64 -> StateT BL.ByteString Maybe BL.ByteString
 parseBytes n = StateT $ \bl -> do
-  let (before, after) = BL.splitAt n bl
-  guard (BL.length before == n)
-  pure (before, after)
+  let (prefix, suffix) = BL.splitAt n bl
+  guard (BL.length prefix == n)
+  pure (prefix, suffix)
 
 -- | Parses a big-endian 64-bit unsigned integer.
 parseWord64BE :: StateT BL.ByteString Maybe Word64
@@ -268,7 +321,7 @@ parseWhile p = StateT (Just . BL.span p)
 -- | Run-length encode lazy a 'BL.ByteString'
 -- for concise display in test results.
 rle :: BL.ByteString -> [(Int, Word8)]
-rle = map (length &&& head) . group . BL.unpack
+rle = map (NE.length &&& NE.head) . NE.group . BL.unpack
 
 -- | Please adjust this expected size of the metadata header
 -- to match that expected of the current implementation.
